/* Pinout of the GB extension:
   BUFFER (8)			NOT USED
   B_8255 PORT A (8)	NOT USED
   B_8255 PORT B (8)	NOT USED
   B_8255 PORT Cl (4)	bit 0: /WR
						bit 1: /RD
						bit 2: /MREQ
   B_8255 PORT Ch (4)	NOT USED
   A_8255 PORT A (8)	ADDR 0..7
   A_8255 PORT B (8)	ADDR 8..15
   A_8255 PORT Cl (4)	DATA 0..3
   A_8255 PORT Ch (4)	DATA 4..7
   GND,VCC... (8)
 */

#ifdef NAME
#undef NAME
#endif
#define NAME "GB"

#define GB_WR	0
#define GB_RD	1
#define GB_MREQ 2

#define GB_IDLE		OUTPORT(B_8255_PC, 0xFF)
#define GB_RD_ROM	OUTPORT(B_8255_PROG, CLR_PC(GB_RD))

#define GB_NO_MBC	0
#define GB_MBC1		1
#define GB_MBC2		2

#define GB_NO_SRAM	0

#define GB_ROM		0
#define GB_SRAM	1

char *gb_type[] = {
	"ROM ONLY", "ROM+MBC1", "ROM+MBC1+RAM", "ROM+MBC1+RAM+BATTERY",
	"UNKNOWN", "ROM+MBC2", "ROM+MBC2+BATTERY"
};
int gb_rom[] = { 32, 64, 128, 256, 512 };
int gb_ram[] = { 0, 2, 8, 32 };
int gb_mbc[] = { GB_NO_MBC, GB_MBC1, GB_MBC1, GB_MBC1, GB_MBC2, GB_MBC2, GB_MBC2 };
int gb_sram[] = { GB_NO_SRAM, GB_NO_SRAM, GB_NO_SRAM, GB_MBC1, GB_NO_SRAM, GB_NO_SRAM, GB_MBC2 };

unsigned char *gb_header = NULL;

INLINE int gb_read_byte(register unsigned int addr)
{
	OUTPORT(A_8255_PB, (addr>>8)&0xFF);
	OUTPORT(A_8255_PA, addr&0xFF);
	return INPORT(A_8255_PC);
}

INLINE void gb_write_byte(register int val, register unsigned int addr)
{
	GB_IDLE;
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_out);
	OUTPORT(A_8255_PC, val);
	OUTPORT(A_8255_PB, (addr>>8)&0xFF);
	OUTPORT(A_8255_PA, addr&0xFF);
	OUTPORT(B_8255_PROG, CLR_PC(GB_MREQ));
	OUTPORT(B_8255_PROG, CLR_PC(GB_WR));
	OUTPORT(B_8255_PROG, SET_PC(GB_WR));
	OUTPORT(B_8255_PROG, SET_PC(GB_MREQ));
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_in|PCh_in);
	GB_RD_ROM;
}

INLINE void gb_write_bksw(register int val, register unsigned int addr)
{
	GB_IDLE;
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_out);
	OUTPORT(A_8255_PC, val);
	OUTPORT(A_8255_PB, (addr>>8)&0xFF);
	OUTPORT(A_8255_PA, addr&0xFF);
	OUTPORT(B_8255_PROG, CLR_PC(GB_WR));
	OUTPORT(B_8255_PROG, SET_PC(GB_WR));
 	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_in|PCh_in);
	GB_RD_ROM;
}

void gb_switch_init()
{
	/*
	 * It seems that the bank-switch IC requires to be written the value
	 * 0x0A at address 0x0000 before being usable...
	 */
	gb_write_bksw(0x0A, 0x0000);
}

void gb_switch_bank(int bank, int type)
{
	GB_IDLE;
	if(type == GB_ROM)
		gb_write_bksw(bank, 0x2100);
	else if(type == GB_SRAM)
		gb_write_bksw(bank, 0x4000);
}

void gb_read_header()
{
	unsigned int addr;

	GB_RD_ROM;
	if(gb_header != NULL)
		return;

	gb_header = (unsigned char *)malloc(0x50 * sizeof(unsigned char));

	for(addr = 0x0100; addr < 0x0150; addr++)
		gb_header[addr-0x0100] = gb_read_byte(addr);
}

void gb_init()
{
	check_extension(NAME);

	/* Configure 8255's */
	/* B:
			port A: INPUT
			port B: INPUT
			port Cl: OUTPUT	  0: /WR
							  1: /RD
							  2: /MREQ
			port Ch: OUTPUT
	   A:
			port A: OUTPUT	  ADDR 0..7
			port B: OUTPUT	  ADDR 8..15
			port Cl: INPUT	  DATA 0..3
			port Ch: INPUT	  DATA 4..7
	*/
	OUTPORT(B_8255_PROG, 0x80|PA_in|PB_in|PCl_out|PCh_out);
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_in|PCh_in);
	GB_IDLE;
}

void gb_analyze()
{
	gb_read_header();

	printf("Game name: %s\n", &gb_header[0x34]);
	printf("Cartridge type: %d(%s)\n", gb_header[0x47], gb_type[gb_header[0x47]]);
	printf("ROM size: %d(%d kB)\n", gb_header[0x48], gb_rom[gb_header[0x48]]);
	if(gb_sram[gb_header[0x49]] == GB_MBC2)
		printf("RAM size: 256 B\n");
	else
		printf("RAM size: %d(%d kB)\n", gb_header[0x49], gb_ram[gb_header[0x49]]);
}

void gb_save(FILE *fp)
{
	int bank, nbbank;
	int sizekB;
	unsigned char val;
	unsigned int addr, min, max;
	unsigned int chk, checksum;

	gb_analyze();

	sizekB = gb_rom[gb_header[0x48]];

	/* One bank is 16kB */
	nbbank = sizekB / 16;
	chk = 0;

	for(bank = 0; bank < nbbank; bank++) {
		printf("Reading bank %d\n", bank);
		gb_switch_bank(bank, GB_ROM);
		min = (bank ? 0x4000 : 0x0000);
		max = min + 0x4000;
		for(addr = min; addr < max; addr++) {
			if(addr % 0x100 == 0) {
				putchar('.');
				fflush(stdout);
			}
			val = gb_read_byte(addr);
			fputc(val, fp);
			chk += val;
		}
		putchar('\n');
	}
	chk -= gb_header[0x4E];
	chk -= gb_header[0x4F];
	chk &= 0xFFFF;
	checksum = ((unsigned int)gb_header[0x4E] << 8) + gb_header[0x4F];;
	if(chk == checksum)
		printf("Checksum OK\n");
	else
		printf("Warning: checksum is wrong (waiting for %X, got %X)\n", checksum, chk);
}

void gb_backup(FILE *fp)
{
	int bank, nbbank, banksize;
	int type, sizekB;
	unsigned int addr, max;

	gb_analyze();

	type = gb_sram[gb_header[0x47]];
	sizekB = gb_ram[gb_header[0x49]];

	if(type == GB_MBC1) {
		/* One bank is 8kB */
		nbbank = sizekB / 8;
		if(nbbank == 0) {
			nbbank = 1;
			banksize = sizekB << 2;
		} else
			banksize = 0x20;
	} else if(type == GB_MBC2) {
		/* SRAM is 512 * 4 bits */
		nbbank = 1;
		banksize = 0x02;
	} else {
		printf("No save-RAM\n");
		return;
	}
	max = 0xA000 + (banksize<<8);

	/* Initialize the bank-switch IC */
	gb_switch_init();
	for(bank = 0; bank < nbbank; bank++) {
		printf("Reading bank %d\n", bank);
		if(type == GB_MBC1) {
			gb_switch_bank(bank, GB_SRAM);
		}
		for(addr = 0xA000; addr < max; addr++) {
			if(addr % 0x100 == 0) {
				putchar('.');
				fflush(stdout);
			}
			fputc(gb_read_byte(addr), fp);
		}
		putchar('\n');
	}
}

void gb_restore(FILE *fp)
{
	int bank, nbbank, banksize;
	int type, sizekB;
	unsigned char val;
	unsigned int addr, max;

	gb_analyze();

	type = gb_sram[gb_header[0x47]];
	sizekB = gb_ram[gb_header[0x49]];

	if(type == GB_MBC1) {
		/* One bank is 8kB */
		nbbank = sizekB / 8;
		if(nbbank == 0) {
			nbbank = 1;
			banksize = sizekB << 2;
		} else
			banksize = 0x20;
	} else if(type == GB_MBC2) {
		/* SRAM is 512 * 4 bits */
		nbbank = 1;
		banksize = 0x02;
	} else {
		printf("No save-RAM\n");
		return;
	}
	max = 0xA000 + (banksize<<8);

	/* Initialize the bank-switch IC */
	gb_switch_init();
	for(bank = 0; bank < nbbank; bank++) {
		printf("Restoring bank %d\n", bank);
		if(type == GB_MBC1) {
			gb_switch_bank(bank, GB_SRAM);
		}
		for(addr = 0xA000; addr < max; addr++) {
			if(feof(fp)) {
				printf("Unexpected EOF\n");
				exit(1);
			}
			if(addr % 0x100 == 0) {
				putchar('.');
				fflush(stdout);
			}
			val = fgetc(fp);
			gb_write_byte(val, addr);
		}
		putchar('\n');
	}
}

void gb_clear()
{
	int bank, nbbank, banksize;
	int type, sizekB;
	unsigned int addr, max;

	gb_analyze();

	type = gb_sram[gb_header[0x47]];
	sizekB = gb_ram[gb_header[0x49]];

	if(type == GB_MBC1) {
		/* One bank is 8kB */
		nbbank = sizekB / 8;
		if(nbbank == 0) {
			nbbank = 1;
			banksize = sizekB << 2;
		} else
			banksize = 0x20;
	} else if(type == GB_MBC2) {
		/* SRAM is 512 * 4 bits */
		nbbank = 1;
		banksize = 0x02;
	} else {
		printf("No save-RAM\n");
		return;
	}
	max = 0xA000 + (banksize<<8);

	/* Initialize the bank-switch IC */
	gb_switch_init();
	for(bank = 0; bank < nbbank; bank++) {
		printf("Clearing bank %d\n", bank);
		if(type == GB_MBC1) {
			gb_switch_bank(bank, GB_SRAM);
		}
		for(addr = 0xA000; addr < max; addr++) {
			if(addr % 0x100 == 0) {
				putchar('.');
				fflush(stdout);
			}
			gb_write_byte(0x00, addr);
		}
		putchar('\n');
	}
}

void gb_test()
{
	int bank, nbbank, banksize;
	int type, sizekB;
	int i, test_val;
	unsigned int addr, max;

	gb_analyze();

	type = gb_sram[gb_header[0x47]];
	sizekB = gb_ram[gb_header[0x49]];

	if(type == GB_MBC1) {
		/* One bank is 8kB */
		nbbank = sizekB / 8;
		if(nbbank == 0) {
			nbbank = 1;
			banksize = sizekB << 2;
		} else
			banksize = 0x20;
	} else if(type == GB_MBC2) {
		/* SRAM is 512 * 4 bits */
		nbbank = 1;
		banksize = 0x02;
	} else {
		printf("No save-RAM\n");
		return;
	}
	max = 0xA000 + (banksize<<8);

	/* Initialize the bank-switch IC */
	gb_switch_init();
	for(bank = 0; bank < nbbank; bank++) {
		int j = (type == GB_MBC1 ? 0xFF : 0x0F);
		printf("Testing bank %d\n", bank);
		if(type == GB_MBC1) {
			gb_switch_bank(bank, GB_SRAM);
		}
		for(i = 0; test_patterns[i] >= 0; i++) {
			test_val = test_patterns[i] & j;
			printf("Writing 0x%02X's\n", test_val);
			for(addr = 0xA000; addr < max; addr++) {
				if(addr % 0x100 == 0) {
					putchar('.');
					fflush(stdout);
				}
				gb_write_byte(test_val, addr);
			}
			putchar('\n');
			printf("Verifying 0x%02X's\n", test_val);
			for(addr = 0xA000; addr < max; addr++) {
				if(addr % 0x100 == 0) {
					putchar('.');
					fflush(stdout);
				}
				if((gb_read_byte(addr)|j) != test_val) {
					fprintf(stderr, "\nError at addr 0x%04X\n", addr);
					exit(-1);
				}
			}
			putchar('\n');
		}
	}
	printf("OK\n");
}

Extension gb_ext = {
	NAME,
	gb_init,
	NULL,		/* no_flag */
	gb_analyze,
	gb_save,
	gb_backup,
	gb_restore,
	gb_clear,
	gb_test,
	NULL,		/* debug */
	NULL,		/* extra */
	NULL		/* usage */
};
