/* Pinout of the NES extension:
   BUFFER (8)           NOT USED
   B_8255 PORT A (8)    NOT USED
   B_8255 PORT B (8)    CHR DATA 0..7
   B_8255 PORT Cl (4)   bit 0: /CHR RD
						bit 1: /CHR WR
						bit 2: /PRG OEN
						bit 3: R/W
   B_8255 PORT Ch (4)   bit 0: CLK
						bit 1..3: NOT USED
   A_8255 PORT A (8)    PRG/CHR ADDR 0..7
   A_8255 PORT B (8)    PRG/CHR ADDR 8..15
   A_8255 PORT Cl (4)   PRG DATA 0..3
   A_8255 PORT Ch (4)   PRG DATA 4..7
   GND,VCC... (8)
 */

#ifdef NAME
#undef NAME
#endif
#define NAME "NES"

#define NES_CRD	 	0
#define NES_CWR		1
#define NES_POE		2
#define NES_RW		3
#define NES_CLK		4

#define NES_PRG		0
#define NES_CHR		1

#define NES_IDLE		OUTPORT(B_8255_PC, 0xFF)
#define NES_READ_PRG	OUTPORT(B_8255_PC, 0xFF-(0x01<<NES_POE))
#define NES_READ_CHR	OUTPORT(B_8255_PC, 0xFF-(0x01<<NES_CRD))

#define NES_WR_CHR(val, addr) \
	OUTPORT(B_8255_PC, 0xFF); \
	OUTPORT(B_8255_PROG, 0x80|PA_in|PB_out|PCl_out|PCh_out); \
	OUTPORT(B_8255_PC, 0xFF); \
	OUTPORT(A_8255_PB, ((addr)>>8)&0xFF); \
	OUTPORT(A_8255_PA, (addr)&0xFF); \
	OUTPORT(B_8255_PB, val); \
	OUTPORT(B_8255_PROG, CLR_PC(NES_CWR)); \
	OUTPORT(B_8255_PROG, SET_PC(NES_CWR)); \
	OUTPORT(B_8255_PROG, 0x80|PA_in|PB_in|PCl_out|PCh_out); \
	OUTPORT(B_8255_PC, 0xFF)

#define NES_WR_PRG(val, addr) \
	OUTPORT(B_8255_PC, 0xFF-(0x01<<NES_CLK)); \
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_out); \
	OUTPORT(A_8255_PB, ((addr)>>8)&0xFF); \
	OUTPORT(A_8255_PA, (addr)&0xFF); \
	OUTPORT(A_8255_PC, val); \
	OUTPORT(B_8255_PROG, CLR_PC(NES_RW)); \
	OUTPORT(B_8255_PROG, SET_PC(NES_CLK)); \
	OUTPORT(B_8255_PROG, CLR_PC(NES_POE)); \
	OUTPORT(B_8255_PROG, CLR_PC(NES_CLK)); \
	OUTPORT(B_8255_PROG, SET_PC(NES_POE)); \
	OUTPORT(B_8255_PROG, SET_PC(NES_RW)); \
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_in|PCh_in); \
	OUTPORT(B_8255_PROG, SET_PC(NES_CLK))

#define NES_RAW_FORMAT	0
#define NES_INES_FORMAT	1

#define MAXBANKS    0x100

int nes_verify = 0;
int nes_format = NES_INES_FORMAT;

typedef struct Nes_info_ {
	int type;
	int nb_prg;
	int prg_bank_size;
	int nb_chr;
	int chr_bank_size;
	int has_vram;
} Nes_info;

Nes_info nes = {
	-1,
	-1,
	-1,
	-1,
	-1,
	0
};

int nes_types[] = {	/* Known types */
	1,
	2,
	3,
	4,
	7,
	-1
};

INLINE int nes_read_byte(register unsigned int addr, int mem)
{
	register int i;

	OUTPORT(A_8255_PB, addr>>8);
	OUTPORT(A_8255_PA, addr&0xFF);
	i = INPORT(mem == NES_PRG ? A_8255_PC : B_8255_PB);
	if(nes_verify >= 2) {
		DELAY;
		if(i != INPORT(mem == NES_PRG ? A_8255_PC : B_8255_PB)) {
			fprintf(stderr, "\nError reading addr 0x%04X\n", addr);
			exit(1);
		}
	}
	return i;
}

unsigned int nes_checksum(unsigned int start, unsigned int length, int mem)
{
	unsigned int chk = 0;
	unsigned int i, max;

	if(mem == NES_PRG) {
		NES_READ_PRG;
	} else if(mem == NES_CHR) {
		NES_READ_CHR;
	}

	max = start+length;
	for(i = start; i < max; i++)
		chk += nes_read_byte(i, mem);
	if(nes_verify >= 1) {
		unsigned int chk2 = 0;
		for(i = start; i < max; i++)
			chk2 += nes_read_byte(i, mem);
		if(chk != chk2) {
			fprintf(stderr, "\nError while verifying checksum: "
				"1st checksum = 0x%04X - "
				"2nd checksum = 0x%04X\n", chk, chk2);
			exit(1);
		}
	}
	return chk;
}

void nes_switch_init()
{
	switch(nes.type) {
	case 1:
		/* Reset registers */
		NES_WR_PRG(0x87, 0x9FFF);
		NES_WR_PRG(0x00, 0xBFFF);
		NES_WR_PRG(0x21, 0xDFFF);
		NES_WR_PRG(0xC8, 0xFFFF);
		/* Set CHR pages to 4kB */
		NES_WR_PRG(0x1E, 0x9FFF);
		NES_WR_PRG(0x0F, 0x9FFF);
		NES_WR_PRG(0x07, 0x9FFF);
		NES_WR_PRG(0x03, 0x9FFF);
		NES_WR_PRG(0x01, 0x9FFF);
		break;
	}
}

void nes_switch_bank(int bank, int mem)
{
	switch(nes.type) {
	case 1:
		/* 16kB PRG banks switched at 0x8000
		 *
		 * 8kB of CHR banks switched at 0x0000
		 *  or
		 * 2x4kB of CHR banks switched separately at 0x0000 and 0x1000
		 */
		if(mem == NES_PRG) {
			NES_WR_PRG(bank, 0xFFFF);
			NES_WR_PRG(bank>>1, 0xFFFF);
			NES_WR_PRG(bank>>2, 0xFFFF);
			NES_WR_PRG(bank>>3, 0xFFFF);
			NES_WR_PRG(bank>>4, 0xFFFF);
		} else if(mem == NES_CHR) {
			NES_WR_PRG(bank, 0xBFFF);
			NES_WR_PRG(bank>>1, 0xBFFF);
			NES_WR_PRG(bank>>2, 0xBFFF);
			NES_WR_PRG(bank>>3, 0xBFFF);
			NES_WR_PRG(bank>>4, 0xBFFF);
			NES_WR_PRG(bank, 0xDFFF);
			NES_WR_PRG(bank>>1, 0xDFFF);
			NES_WR_PRG(bank>>2, 0xDFFF);
			NES_WR_PRG(bank>>3, 0xDFFF);
			NES_WR_PRG(bank>>4, 0xDFFF);
		}
		break;
	case 2:
		/* 16kB PRG banks switched at 0x8000
		 *
		 * No CHR switch
		 */
		if(mem == NES_PRG) {
			unsigned int addr;
			/* Find address where value == bank (avoid bus conflict) */
			NES_READ_PRG;
			for(addr = 0xC000; addr <= 0xFFFF; addr++) {
				if(nes_read_byte(addr, mem) == bank) {
					printf("%04X <- %d\n", addr, bank);
					NES_WR_PRG(bank, addr);
					break;
				}
			}
		}
		break;
	case 3:
		/* No PRG switch
		 *
		 * 8kB of CHR banks switched at 0x0000
		 */
		if(mem == NES_CHR) {
			NES_WR_PRG(bank, 0x8000);
		}
		break;
	case 4:
		/* 2x8kB PRG banks switched at 0x8000 and 0xA000
		 *
		 * 2x2kB of CHR banks switched separately at 0x0000 and 0x0800
		 *  or
		 * 4x1kB of CHR banks switched separately at 0x0000, 0x0400, 0x0800, and 0x0C00
		 */
		if(mem == NES_PRG) {
			NES_WR_PRG(6, 0x8000);
			NES_WR_PRG(bank<<1, 0x8000+1);
			NES_WR_PRG(7, 0x8000);
			NES_WR_PRG((bank<<1)+1, 0x8000+1);
		} else if(mem == NES_CHR) {
			NES_WR_PRG(0x00, 0x8000);
			NES_WR_PRG(bank<<3, 0x8000+1);
			NES_WR_PRG(0x01, 0x8000);
			NES_WR_PRG((bank<<3)+2, 0x8000+1);
			NES_WR_PRG(0x02, 0x8000);
			NES_WR_PRG((bank<<3)+4, 0x8000+1);
			NES_WR_PRG(0x03, 0x8000);
			NES_WR_PRG((bank<<3)+5, 0x8000+1);
			NES_WR_PRG(0x04, 0x8000);
			NES_WR_PRG((bank<<3)+6, 0x8000+1);
			NES_WR_PRG(0x05, 0x8000);
			NES_WR_PRG((bank<<3)+7, 0x8000+1);
		}
		break;
	case 7:
		/* 32kB PRG banks switched at 0x8000
		 *
		 * No CHR switch
		 */
		if(mem == NES_PRG) {
			NES_WR_PRG(bank, 0x8000+bank);
		}
		break;
	}
	if(mem == NES_PRG) {
		NES_READ_PRG;
	} else if(mem == NES_CHR) {
		NES_READ_CHR;
	}
}

int nes_check_type()
{
	unsigned int chk1, chk2;

	/* Calculate PRG checksums on 4kB */
	/* Calculate CHR checksums on 1kB */

	printf("Testing MMC %d\n", nes.type);
	switch(nes.type) {
	case 1:
	case 2:
	case 4:
		/* Check PRG 0x8000 -> 0xBFFF (should be switched) */
		nes_switch_bank(0, NES_PRG);
		chk1 = nes_checksum(0x0000, 0x1000, NES_PRG);
		nes_switch_bank(1, NES_PRG);
		chk2 = nes_checksum(0x0000, 0x1000, NES_PRG);
		if(chk1 == chk2)
			return 0;
		/* Check 0xC000 -> 0xFFFF (should not be switched) */
		nes_switch_bank(0, NES_PRG);
		chk1 = nes_checksum(0x4000, 0x1000, NES_PRG);
		nes_switch_bank(1, NES_PRG);
		chk2 = nes_checksum(0x4000, 0x1000, NES_PRG);
		if(chk1 != chk2)
			return 0;
		return 1;
	case 3:
		/* Check CHR 0x0000 -> 0x1FFF (should be switched) */
		nes_switch_bank(0, NES_CHR);
		chk1 = nes_checksum(0x0000, 0x0400, NES_CHR);
		nes_switch_bank(1, NES_CHR);
		chk2 = nes_checksum(0x0000, 0x0400, NES_CHR);
		if(chk1 == chk2)
			return 0;
		return 1;
	case 7:
		/* Check PRG 0x8000 -> 0xBFFF (should be switched) */
		nes_switch_bank(0, NES_PRG);
		chk1 = nes_checksum(0x0000, 0x1000, NES_PRG);
		nes_switch_bank(1, NES_PRG);
		chk2 = nes_checksum(0x0000, 0x1000, NES_PRG);
		if(chk1 == chk2)
			return 0;
		/* Check 0xC000 -> 0xFFFF (should be switched) */
		nes_switch_bank(0, NES_PRG);
		chk1 = nes_checksum(0x4000, 0x1000, NES_PRG);
		nes_switch_bank(1, NES_PRG);
		chk2 = nes_checksum(0x4000, 0x1000, NES_PRG);
		if(chk1 == chk2)
			return 0;
		return 1;
	}
	return 0;
}

void nes_find_type()
{
	int i;

	for(i = 0; nes_types[i] != -1; i++) {
		nes.type = nes_types[i];
		nes_switch_init();
		if(nes_check_type())
			return;
	}
	nes.type = 0;
}

void nes_count_prg_banks()
{
	int bank;
	unsigned int first, last, chk;

	if(nes.type == 0) {
		nes.nb_prg = 1;
		return;
	}
	for(bank = 0; bank < MAXBANKS; bank++) {
		nes_switch_bank(bank, NES_PRG);
		printf("Testing PRG bank %d...", bank);
		fflush(stdout);
		chk = nes_checksum(0x0000, 0x4000, NES_PRG);
		printf(" (%04X)\n", chk);
		if(bank == 0)
			first = last = chk;
		else {
			if(chk == first || chk == last) {
				nes.nb_prg = bank;
				break;
			}
			last = chk;
		}
	}
	if(bank == MAXBANKS)
		nes.nb_prg = 0;
}

void nes_count_chr_banks()
{
	int bank;
	unsigned int first, last, chk;

	if(nes.type == 0) {
		nes.nb_chr = 1;
		return;
	}

	for(bank = 0; bank < MAXBANKS; bank++) {
		nes_switch_bank(bank, NES_CHR);
		printf("Testing CHR bank %d...", bank);
		fflush(stdout);
		chk = nes_checksum(0x0000, 0x2000, NES_CHR);
		printf(" (%04X)\n", chk);
		if(bank == 0)
			first = last = chk;
		else {
			if(chk == first || chk == last) {
				nes.nb_chr = bank;
				break;
			}
			last = chk;
		}
	}
	if(bank == MAXBANKS)
		nes.nb_chr = 0;
}

void nes_test_vram()
{
	int i;

	nes_switch_bank(0, NES_CHR);

	OUTPORT(A_8255_PB, 0x00);
	OUTPORT(A_8255_PA, 0x00);
	i = INPORT(B_8255_PB);

	NES_WR_CHR(0x00, 0x0000);
	if(INPORT(B_8255_PB) != 0x00) {
		nes.has_vram = 0;
		return;
	}

	NES_WR_CHR(0xFF, 0x0000);
	if(INPORT(B_8255_PB) != 0xFF) {
		nes.has_vram = 0;
		return;
	}

	NES_WR_CHR(i, 0x0000);
	nes.has_vram = 1;
}

void nes_find_prg_bank_size()
{
	/* 16kB or 32kB */
	if(nes.type == 0) {
		unsigned int chk1, chk2;
		chk1 = nes_checksum(0x0000, 0x4000, NES_PRG);
		chk2 = nes_checksum(0x4000, 0x4000, NES_PRG);
		if(chk1 == chk2)
			nes.prg_bank_size = 16;
		else
			nes.prg_bank_size = 32;
	} else if(nes.type == 1) {
		/* Some games only use the MMC for CHR, and have 1*32kB PRG */
		if(nes.nb_prg == 1) {
			unsigned int chk1, chk2;
			chk1 = nes_checksum(0x0000, 0x4000, NES_PRG);
			chk2 = nes_checksum(0x4000, 0x4000, NES_PRG);
			if(chk1 == chk2)
				nes.prg_bank_size = 16;
			else
				nes.prg_bank_size = 32;
		} else
			nes.prg_bank_size = 16;
	} else if(nes.type == 7)
		nes.prg_bank_size = 32;
	else
		nes.prg_bank_size = 16;
}

void nes_find_chr_bank_size()
{
	/* 4kB or 8kB */
	if(nes.type == 1)
		nes.chr_bank_size = 4;
	else
		nes.chr_bank_size = 8;
}

void nes_read_info() {
	if(nes.type < 0) {
		printf(
		"Automatic MMC type detection is not yet available\n"
		"Use the '-m' flag to specify cartridge MMC type\n");
		exit(1);
		nes_find_type();
	} else {
		nes_switch_init();
	}
	if(nes.nb_prg < 0) {
		nes_count_prg_banks();
	}
	if(nes.prg_bank_size < 0) {
		nes_find_prg_bank_size();
	}
	if(nes.nb_chr < 0) {
		nes_count_chr_banks();
	}
	if(nes.chr_bank_size < 0) {
		nes_find_chr_bank_size();
	}
	if(nes.has_vram < 0) {
		nes_test_vram();
	}
}

void nes_analyze()
{
	nes_read_info();

	printf("MBC type: %d\n", nes.type);
	printf("Number of PRG banks: %d\n", nes.nb_prg);
	printf("Size of PRG banks: %d kB\n", nes.prg_bank_size);
	printf("Number of CHR banks: %d\n", nes.nb_chr);
	printf("Size of CHR banks: %d kB\n", nes.chr_bank_size);
	printf("Type of CHR banks: %s\n", nes.has_vram ? "VRAM" : "VROM");
}

void nes_read_prg(FILE *fp)
{
	int bank, i;
	unsigned int addr, chk;
	int banksize;

	banksize = nes.prg_bank_size<<10;

	for(bank = 0; bank < nes.nb_prg; bank++) {
		printf("Reading bank %d\n", bank);
		nes_switch_bank(bank, NES_PRG);
		chk = 0;
		for(addr = 0x0000; addr < banksize; addr++) {
			if((addr&0xFF) == 0) {
				putchar('.');
				fflush(stdout);
			}
			i = nes_read_byte(addr, NES_PRG);
			chk += i;
			fputc(i, fp);
		}
		printf(" (%04X)\n", chk);
		if(nes_verify >= 1) {
			unsigned int chk2 = 0;
			printf("Verifying bank %d\n", bank);
			for(addr = 0x0000; addr < banksize; addr++) {
				if((addr&0xFF) == 0) {
					putchar('.');
					fflush(stdout);
				}
				chk2 += nes_read_byte(addr, NES_PRG);
			}
			printf(" (%04X)\n", chk2);
			if(chk != chk2) {
				fprintf(stderr, "\nError while verifying: "
					"1st checksum = 0x%04X - "
					"2nd checksum = 0x%04X\n", chk, chk2);
				exit(1);
			}
		}
	}
}

void nes_read_chr(FILE *fp)
{
	int bank, i;
	unsigned int addr, chk;
	int banksize;

	banksize = nes.chr_bank_size<<10;

	for(bank = 0; bank < nes.nb_chr; bank++) {
		printf("Reading bank %d\n", bank);
		nes_switch_bank(bank, NES_CHR);
		chk = 0;
		for(addr = 0x0000; addr < banksize; addr++) {
			if((addr&0xFF) == 0) {
				putchar('.');
				fflush(stdout);
			}
			i = nes_read_byte(addr, NES_CHR);
			chk += i;
			fputc(i, fp);
		}
		printf(" (%04X)\n", chk);
		if(nes_verify >= 1) {
			unsigned int chk2 = 0;
			printf("Verifying bank %d\n", bank);
			for(addr = 0x0000; addr < banksize; addr++) {
				if((addr&0xFF) == 0) {
					putchar('.');
					fflush(stdout);
				}
				chk2 += nes_read_byte(addr, NES_CHR);
			}
			printf(" (%04X)\n", chk2);
			if(chk != chk2) {
				fprintf(stderr, "\nError while verifying: "
					"1st checksum = 0x%04X - "
					"2nd checksum = 0x%04X\n", chk, chk2);
				exit(1);
			}
		}
	}
}

void nes_save(FILE *fp)
{
	nes_analyze();

	if(nes_format == NES_RAW_FORMAT) {
		nes_read_prg(fp);
		if(!nes.has_vram)
			nes_read_chr(fp);
	} else if(nes_format == NES_INES_FORMAT) {
		unsigned char hdr[16];

		memset(hdr, 0, 16);
		strcpy(hdr, "NES\032");
		/* iNES: always 16kB PRG blocks */
		hdr[4] = (unsigned char)((nes.nb_prg*nes.prg_bank_size)>>4);
		/* iNES: always 8kB CHR blocks */
		if(nes.has_vram)
			hdr[5] = 0;
		else
			hdr[5] = (unsigned char)((nes.nb_chr*nes.chr_bank_size)>>3);
		hdr[6] = (unsigned char)(nes.type<<4);
		fwrite(hdr, 16, 1, fp);
		nes_read_prg(fp);
		if(!nes.has_vram)
			nes_read_chr(fp);
	}
}

void nes_init()
{
	check_extension(NAME);

	/* Configure 8255's */
	/* B:
			port A: INPUT     NOT USED
			port B: INPUT     CHR DATA 0..7
			port Cl: OUTPUT   0: /CHR RD
							  1: /CHR WR
							  2: /PRG OEN
							  3: R/W
			port Ch: OUTPUT   0: CLK
							  1..3: NOT USED
	   A:
			port A: OUTPUT    PRG/CHR ADDR 0..7
			port B: OUTPUT    PRG/CHR ADDR 8..15
			port Cl: INPUT    PRG DATA 0..3
			port Ch: INPUT    PRG DATA 4..7
	*/
	OUTPORT(B_8255_PROG, 0x80|PA_in|PB_in|PCl_out|PCh_out);
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_in|PCh_in);
	NES_IDLE;
}

void nes_debug()
{
	char s[32], file[16];
	unsigned int addr, chk, len;
	int val;
	FILE *fp;

	while(1) {
		printf("\n[Command,'?']-> ");
		fflush(stdout);
		fflush(stdin);

		fgets(s, 31, stdin);
		for(val = 0; s[val] >= ' '; val++)
			s[val] = toupper(s[val]);
		s[val]='\0';

		switch(s[0]) {
			case 'H':
			case '?':
				puts("RP<addr>       : Read PRG <addr> (hex)");
				puts("WP<addr> <val> : Write PRG <val> (hex) to <addr> (hex)");
				puts("RC<addr>       : Read CHR <addr> (hex)");
				puts("WC<addr> <val> : Write CHR <val> to <addr> (hex)");
				puts("CP             : Calculate checksum of PRG bank");
				puts("CC             : Calculate checksum of CHR bank");
				puts("SP<len> <file> : Save <len>kB (dec) of PRG into <file>");
				puts("SC             : Save <len>kB (dec) of CHR into <file>");
				puts("?,H            : Show this help text");
				puts("Q              : Exit");
				break;
			case 'R':
				sscanf(s+2, "%X", &addr);
				OUTPORT(A_8255_PB, addr>>8);
				OUTPORT(A_8255_PA, addr&0xFF);
				if(s[1] == 'P') {
					NES_READ_PRG;
					val = INPORT(A_8255_PC);
					printf("PRG value at %04X is %02X\n", addr, val);
				} else if(s[1] == 'C') {
					NES_READ_CHR;
					val = INPORT(B_8255_PB);
					printf("CHR value at %04X is %02X\n", addr, val);
				}
				break;
			case 'W':
				sscanf(s+2, "%X %X", &addr, &val);
				if(s[1] == 'P') {
					NES_WR_PRG(val, addr);
				} else if(s[1] == 'C') {
					NES_WR_CHR(val, addr);
				}
				break;
			case 'C':
				if(s[1] == 'P') {
					/* Calculate checksum on 32kB */
					printf("PRG checksum is %04X %04X %04X %04X (4x8kB)\n",
					nes_checksum(0x0000, 0x2000, NES_PRG),
					nes_checksum(0x2000, 0x2000, NES_PRG),
					nes_checksum(0x4000, 0x2000, NES_PRG),
					nes_checksum(0x6000, 0x2000, NES_PRG));
				} else if(s[1] == 'C') {
					/* Calculate checksum on 8kB */
					printf("CHR checksum is %04X %04X %04X %04X %04X %04X %04X %04X (8x1kB)\n",
					nes_checksum(0x0000, 0x0400, NES_CHR),
					nes_checksum(0x0400, 0x0400, NES_CHR),
					nes_checksum(0x0800, 0x0400, NES_CHR),
					nes_checksum(0x0C00, 0x0400, NES_CHR),
					nes_checksum(0x1000, 0x0400, NES_CHR),
					nes_checksum(0x1400, 0x0400, NES_CHR),
					nes_checksum(0x1800, 0x0400, NES_CHR),
					nes_checksum(0x1C00, 0x0400, NES_CHR));
				}
				break;
			case 'S':
				if(s[1] != 'P' && s[1] != 'C')
					break;
				sscanf(s+2, "%d %s", &len, file);
				if((fp = fopen(file, "wb")) == NULL) {
					printf("Error opening file %s\n", file);
					exit(1);
				}
				len <<= 10;
				chk = 0;
				printf("Writing 0->0x%X to %s\n", len, file);
				for(addr = 0x0000; addr < len; addr++) {
					if((addr&0xFF) == 0) {
						putchar('.');
						fflush(stdout);
					}
					val = nes_read_byte(addr, (s[1] == 'P' ? NES_PRG : NES_CHR));
					chk += val;
					fputc(val, fp);
				}
				printf(" (%04X)\n", chk);
				fclose(fp);
				break;
			case 'Q':
				exit(0);
		}
	}
}

int nes_extra(char **argv, int argc)
{
	if(argv[0][0] == '-') {
		switch(toupper(argv[0][1])) {
			case 'V':
				nes_verify++;
				return 0;
			case 'W':
				switch(toupper(argv[0][2])) {
					case 'A':
						nes.has_vram = 1;
						break;
					case 'O':
						nes.has_vram = 0;
						break;
					default:
						usage();
				}
				return 0;
			case 'M':
				if(argc < 1)
					usage();
				nes.type = atoi(argv[1]);
				return 1;
			case 'N':
				if(argc < 1)
					usage();
				switch(toupper(argv[0][2])) {
					case 'P':
						nes.nb_prg = atoi(argv[1]);
						break;
					case 'C':
						nes.nb_chr = atoi(argv[1]);
						break;
					default:
						usage();
				}
				return 1;
			case 'L':
				if(argc < 1)
					usage();
				switch(toupper(argv[0][2])) {
					case 'P':
						nes.prg_bank_size = atoi(argv[1]);
						break;
					case 'C':
						nes.chr_bank_size = atoi(argv[1]);
						break;
					default:
						usage();
				}
				return 1;
			case 'F':
				if(argc < 1)
					usage();
				nes_format = atoi(argv[1]);
				return 1;
		}
	}
	return -1;
}

void nes_usage()
{
	printf(
	"NES extension:\n"
	"    -v        Verify (read the cartridge twice)\n"
	"    -v -v     Verify even more\n"
	"    -wa       The cartridge has VRAM\n"
	"    -wo       The cartridge has VROM\n"
	"    -m nb     Specify the memory controler (MMC) of the card\n"
	"    -nc nb    Specify the number of character ROM banks\n"
	"    -np nb    Specify the number of program ROM banks\n"
	"    -lc nb    Specify the length of character ROM banks\n"
	"    -lp nb    Specify the length of program ROM banks\n"
	"    -f nb     Specify format of the save file: 0 = raw, 1 = iNES (default: 1)\n"
	);
}

Extension nes_ext = {
	NAME,
	nes_init,
	NULL,		/* no_flag */
	nes_analyze,
	nes_save,
	NULL,		/* backup */
	NULL,		/* restore */
	NULL,		/* clear */
	NULL,		/* test */
	nes_debug,
	nes_extra,
	nes_usage
};
